>module UsageElim where
>import Util
>import Specification

This optimization removes any tags from the tree that does not affect
the characters under the hierarchy. This is very useful with white
spaces and tags which have no characters inside them. It'll only
look at the contexts attached to characters and context changes
attached to the hierarchy.

    B
    |
    B
    |
    c
becomes:
    B
    |
    c
because topmost B cannot change c at all because of destructive
"assignment" of b-flag. The implementation of this moves usage and
blocking information up and down of the hiearchy and tries to
determine which nodes can be removed. It also uses information
about whitespaces and text to detect whether the output will be
affected by tags.

Limitations of this algorithm is that it can only move information
up or down the hierarchy, never between adjacent nodes.

>data UsageRecordType = URec {b_usage :: Bool,
>			       em_usage :: Bool,
>			       i_usage :: Bool,
>			       s_usage :: Bool,
>			       tt_usage :: Bool,
>			       u_usage :: Bool,
>			       size_usage :: Bool,
>                              color_usage :: Bool } 
>       deriving(Show)

>usageelim :: Rec CTree -> Rec CTree
>usageelim tree = fold checkremovablechilds tree
>
>nousage :: UsageRecordType
>nousage = URec False False False False False False False False

>spaceusage :: UsageRecordType
>spaceusage = URec False False False False True True True True

>fullusage :: UsageRecordType
>fullusage = URec True True True True True True True True


>canweremove :: Rec CTree -> Bool
>canweremove (In tree) = checkremovable tree 

>checkremovable :: CTree (Rec CTree) -> Bool
>checkremovable tree | (nodetype tree) == TagType = let
>         uses = map (fold calcusage) (child tree)
>         taguses = usage (nodecontext tree) (tag tree) nousage
>         mergeuses = foldr mergeusage nousage uses
>         anduses = andusage mergeuses taguses
>       in (not (b_usage anduses || em_usage anduses ||i_usage anduses ||s_usage anduses ||tt_usage anduses ||u_usage anduses ||size_usage anduses ||color_usage anduses))
>checkremovable _ = False

>gencompose :: (Bool -> Bool -> Bool ) -> UsageRecordType -> UsageRecordType -> UsageRecordType
>gencompose f u1 u2 = URec (b_usage u1 `f` b_usage u2)
>			 (em_usage u1 `f` em_usage u2)
>			 (i_usage u1 `f` i_usage u2)
>			 (s_usage u1 `f` s_usage u2)
>			 (tt_usage u1 `f` tt_usage u2)
>			 (u_usage u1 `f` u_usage u2)
>			 (size_usage u1 `f` size_usage u2)
>			 (color_usage u1 `f` color_usage u2)

>mergeusage :: UsageRecordType -> UsageRecordType -> UsageRecordType
>mergeusage = gencompose f
>   where f x y = x ||y

>subtractusage :: UsageRecordType -> UsageRecordType -> UsageRecordType
>subtractusage = gencompose f
>   where f x y = x && not y

>andusage :: UsageRecordType -> UsageRecordType -> UsageRecordType
>andusage = gencompose f
>   where f x y = x && y

>calcusage :: CTree UsageRecordType -> UsageRecordType
>calcusage t1 = aa t1
>   where aa :: CTree UsageRecordType -> UsageRecordType
>         aa t | (nodetype t) == TagType = 
>                       mergeusage (subtractusage 
>                             (foldr mergeusage nousage (child t)) 
>                             (blocksusage (nodecontext t) (tag t) nousage)) 
>                             (tag_uses_prop (tag t))
>           | (nodetype t) == SpaceType && (null (chars t)) = nousage
>           | (nodetype t) == CharType && (null (chars t)) = nousage
>           |(nodetype t) == SpaceType = spaceusage
>           | (nodetype t) == CharType = fullusage
>           | otherwise = error "Odd node type"

>usageelim_forest :: [Rec CTree] -> [Rec CTree]
>usageelim_forest lst = shallow_usageelim_forest (map usageelim lst)

>shallow_usageelim_forest :: [Rec CTree]->[Rec CTree]
>shallow_usageelim_forest lst = let
>	removeBool = map canweremove lst
>	transElimed (True, In (CT (ContextEntry (TagNode _ lst2) _))) = lst2
>	transElimed (True, _) = error "CharNode or SpaceNode cannot be removed"
>	transElimed (False,x) = [x]
>	in concatMap transElimed (zip removeBool lst)

>checkremovablechilds :: CTree (Rec CTree) -> Rec CTree
>checkremovablechilds n | (nodetype n) == TagType = setchilds (shallow_usageelim_forest (child n)) (In n)
>checkremovablechilds x = In x

>tag_uses_prop :: Tag -> UsageRecordType
>tag_uses_prop EMTag = nousage { s_usage = True, em_usage = True }
>tag_uses_prop UTag = nousage { u_usage = True }
>tag_uses_prop _ = nousage

>usage :: Context -> Tag -> UsageRecordType -> UsageRecordType
>usage _ BTag u = u {b_usage = True }
>usage _ ITag u = u { i_usage = True }
>usage _ PLTag u = u { u_usage = True, s_usage = True, em_usage = True, i_usage = True, b_usage = True, tt_usage = True }
>usage _ STag u = u { s_usage = True, em_usage = True }
>usage _ TTTag u = u {tt_usage = True }
>usage _ UTag u = u {u_usage = True }
>usage _ (SizeTag _) u = u { size_usage = True }
>usage _ (ColorTag _) u = u { color_usage = True }
>usage c EMTag u | (not (s_prop c)) = u { em_usage = True, s_usage = True }
>usage _ EMTag u = u

>blocksusage :: Context -> Tag -> UsageRecordType -> UsageRecordType
>blocksusage _ BTag u = u {b_usage = True }
>blocksusage _ ITag u = u { i_usage = True }
>blocksusage _ PLTag u = u { u_usage = True, s_usage = True, em_usage = True, i_usage = True, b_usage = True, tt_usage = True }
>blocksusage _ STag u = u { s_usage = True, em_usage = True }
>blocksusage _ TTTag u = u {tt_usage = True }
>blocksusage _ UTag u = u {u_usage = False }
>blocksusage _ (SizeTag _) u = u { size_usage = True }
>blocksusage _ (ColorTag _) u = u { color_usage = True }
>blocksusage _ EMTag u = u { em_usage = False }




